@@ -74,7 +74,7 @@
 	 | PEER_CAP_ADDPATH_AF_TX_ADV | PEER_CAP_ADDPATH_AF_RX_RCV             \
 	 | PEER_CAP_ENHE_AF_NEGO)
 
-typedef enum { BGP_ATTR_VEC_NH = 0, BGP_ATTR_VEC_MAX } bpacket_attr_vec_type;
+enum bpacket_attr_vec_type { BGP_ATTR_VEC_NH = 0, BGP_ATTR_VEC_MAX };
 
 typedef struct {
 	uint32_t flags;
@@ -109,12 +109,6 @@ struct bpacket {
 struct bpacket_queue {
 	TAILQ_HEAD(pkt_queue, bpacket) pkts;
 
-#if 0
-  /* A dummy packet that is used to thread all peers that have
-     completed their work */
-  struct bpacket sentinel;
-#endif
-
 	unsigned int conf_max_count;
 	unsigned int curr_count;
 	unsigned int hwm_count;
@@ -210,6 +204,9 @@ struct update_subgroup {
 	/* send prefix count */
 	uint32_t scount;
 
+	/* send prefix count prior to packet update */
+	uint32_t pscount;
+
 	/* announcement attribute hash */
 	struct hash *hash;
 
@@ -252,18 +249,20 @@ struct update_subgroup {
 	uint64_t id;
 
 	uint16_t sflags;
-
-	/* Subgroup flags, see below  */
-	uint16_t flags;
-};
-
+#define SUBGRP_STATUS_DEFAULT_ORIGINATE (1 << 0)
+#define SUBGRP_STATUS_FORCE_UPDATES (1 << 1)
+#define SUBGRP_STATUS_TABLE_REPARSING (1 << 2)
 /*
- * We need to do an outbound refresh to get this subgroup into a
- * consistent state.
+ * This flag has been added to ensure that the SNT counters
+ * gets incremented and decremented only during the creation
+ * and deletion workflows of default originate,
+ * not during the update workflow.
  */
-#define SUBGRP_FLAG_NEEDS_REFRESH         (1 << 0)
+#define SUBGRP_STATUS_PEER_DEFAULT_ORIGINATED (1 << 3)
 
-#define SUBGRP_STATUS_DEFAULT_ORIGINATE   (1 << 0)
+	uint16_t flags;
+#define SUBGRP_FLAG_NEEDS_REFRESH (1 << 0)
+};
 
 /*
  * Add the given value to the specified counter on a subgroup and its
@@ -296,10 +295,10 @@ struct updwalk_context {
 	struct bgp_path_info *pi;
 	uint64_t updgrp_id;
 	uint64_t subgrp_id;
-	bgp_policy_type_e policy_type;
+	enum bgp_policy_type policy_type;
 	const char *policy_name;
 	int policy_event_start_flag;
-	int policy_route_update;
+	bool policy_route_update;
 	updgrp_walkcb cb;
 	void *context;
 	uint8_t flags;
@@ -376,14 +375,15 @@ extern void update_subgroup_split_peer(struct peer_af *, struct update_group *);
 extern bool update_subgroup_check_merge(struct update_subgroup *, const char *);
 extern bool update_subgroup_trigger_merge_check(struct update_subgroup *,
 						int force);
-extern void update_group_policy_update(struct bgp *bgp, bgp_policy_type_e ptype,
-				       const char *pname, int route_update,
+extern void update_group_policy_update(struct bgp *bgp,
+				       enum bgp_policy_type ptype,
+				       const char *pname, bool route_update,
 				       int start_event);
 extern void update_group_af_walk(struct bgp *bgp, afi_t afi, safi_t safi,
 				 updgrp_walkcb cb, void *ctx);
 extern void update_group_walk(struct bgp *bgp, updgrp_walkcb cb, void *ctx);
 extern void update_group_periodic_merge(struct bgp *bgp);
-extern int
+extern void
 update_group_refresh_default_originate_route_map(struct thread *thread);
 extern void update_group_start_advtimer(struct bgp *bgp);
 
@@ -417,7 +417,7 @@ extern struct stream *bpacket_reformat_for_peer(struct bpacket *pkt,
 						struct peer_af *paf);
 extern void bpacket_attr_vec_arr_reset(struct bpacket_attr_vec_arr *vecarr);
 extern void bpacket_attr_vec_arr_set_vec(struct bpacket_attr_vec_arr *vecarr,
-					 bpacket_attr_vec_type type,
+					 enum bpacket_attr_vec_type type,
 					 struct stream *s, struct attr *attr);
 extern void subgroup_default_update_packet(struct update_subgroup *subgrp,
 					   struct attr *attr,
@@ -469,7 +469,9 @@ extern int update_group_clear_update_dbg(struct update_group *updgrp,
 					 void *arg);
 
 extern void update_bgp_group_free(struct bgp *bgp);
-extern int bgp_addpath_encode_tx(struct peer *peer, afi_t afi, safi_t safi);
+extern bool bgp_addpath_encode_tx(struct peer *peer, afi_t afi, safi_t safi);
+extern bool bgp_check_selected(struct bgp_path_info *bpi, struct peer *peer,
+			       bool addpath_capable, afi_t afi, safi_t safi);
 
 /*
  * Inline functions
