@@ -42,14 +42,53 @@
 #include "static_zebra.h"
 #include "static_nht.h"
 #include "static_vty.h"
+#include "static_debug.h"
 
-bool debug;
+DEFINE_MTYPE_STATIC(STATIC, STATIC_NHT_DATA, "Static Nexthop tracking data");
+PREDECL_HASH(static_nht_hash);
+
+struct static_nht_data {
+	struct static_nht_hash_item itm;
+
+	struct prefix nh;
+	safi_t safi;
+
+	vrf_id_t nh_vrf_id;
+
+	uint32_t refcount;
+	uint8_t nh_num;
+	bool registered;
+};
+
+static int static_nht_data_cmp(const struct static_nht_data *nhtd1,
+			       const struct static_nht_data *nhtd2)
+{
+	if (nhtd1->nh_vrf_id != nhtd2->nh_vrf_id)
+		return numcmp(nhtd1->nh_vrf_id, nhtd2->nh_vrf_id);
+	if (nhtd1->safi != nhtd2->safi)
+		return numcmp(nhtd1->safi, nhtd2->safi);
+
+	return prefix_cmp(&nhtd1->nh, &nhtd2->nh);
+}
+
+static unsigned int static_nht_data_hash(const struct static_nht_data *nhtd)
+{
+	unsigned int key = 0;
+
+	key = prefix_hash_key(&nhtd->nh);
+	return jhash_2words(nhtd->nh_vrf_id, nhtd->safi, key);
+}
+
+DECLARE_HASH(static_nht_hash, struct static_nht_data, itm, static_nht_data_cmp,
+	     static_nht_data_hash);
+
+static struct static_nht_hash_head static_nht_hash[1];
 
 /* Zebra structure to hold current status. */
 struct zclient *zclient;
-static struct hash *static_nht_hash;
+uint32_t zebra_ecmp_count = MULTIPATH_NUM;
 
-/* Inteface addition message from zebra. */
+/* Interface addition message from zebra. */
 static int static_ifp_create(struct interface *ifp)
 {
 	static_ifindex_update(ifp, true);
@@ -85,12 +124,6 @@ static int interface_address_delete(ZAPI_CALLBACK_ARGS)
 
 static int static_ifp_up(struct interface *ifp)
 {
-	if (if_is_vrf(ifp)) {
-		struct static_vrf *svrf = static_vrf_lookup_by_id(ifp->vrf_id);
-
-		static_fixup_vrf_ids(svrf);
-	}
-
 	/* Install any static reliant on this interface coming up */
 	static_install_intf_nh(ifp);
 	static_ifindex_update(ifp, true);
@@ -110,54 +143,47 @@ static int route_notify_owner(ZAPI_CALLBACK_ARGS)
 	struct prefix p;
 	enum zapi_route_notify_owner note;
 	uint32_t table_id;
-	char buf[PREFIX_STRLEN];
+	safi_t safi;
 
-	if (!zapi_route_notify_decode(zclient->ibuf, &p, &table_id, &note))
+	if (!zapi_route_notify_decode(zclient->ibuf, &p, &table_id, &note, NULL,
+				      &safi))
 		return -1;
 
-	prefix2str(&p, buf, sizeof(buf));
-
 	switch (note) {
 	case ZAPI_ROUTE_FAIL_INSTALL:
-		static_nht_mark_state(&p, vrf_id, STATIC_NOT_INSTALLED);
-		zlog_warn("%s: Route %s failed to install for table: %u",
-			  __func__, buf, table_id);
+		static_nht_mark_state(&p, safi, vrf_id, STATIC_NOT_INSTALLED);
+		zlog_warn("%s: Route %pFX failed to install for table: %u",
+			  __func__, &p, table_id);
 		break;
 	case ZAPI_ROUTE_BETTER_ADMIN_WON:
-		static_nht_mark_state(&p, vrf_id, STATIC_NOT_INSTALLED);
+		static_nht_mark_state(&p, safi, vrf_id, STATIC_NOT_INSTALLED);
 		zlog_warn(
-			"%s: Route %s over-ridden by better route for table: %u",
-			__func__, buf, table_id);
+			"%s: Route %pFX over-ridden by better route for table: %u",
+			__func__, &p, table_id);
 		break;
 	case ZAPI_ROUTE_INSTALLED:
-		static_nht_mark_state(&p, vrf_id, STATIC_INSTALLED);
+		static_nht_mark_state(&p, safi, vrf_id, STATIC_INSTALLED);
 		break;
 	case ZAPI_ROUTE_REMOVED:
-		static_nht_mark_state(&p, vrf_id, STATIC_NOT_INSTALLED);
+		static_nht_mark_state(&p, safi, vrf_id, STATIC_NOT_INSTALLED);
 		break;
 	case ZAPI_ROUTE_REMOVE_FAIL:
-		static_nht_mark_state(&p, vrf_id, STATIC_INSTALLED);
-		zlog_warn("%s: Route %s failure to remove for table: %u",
-			  __func__, buf, table_id);
+		static_nht_mark_state(&p, safi, vrf_id, STATIC_INSTALLED);
+		zlog_warn("%s: Route %pFX failure to remove for table: %u",
+			  __func__, &p, table_id);
 		break;
 	}
 
 	return 0;
 }
+
 static void zebra_connected(struct zclient *zclient)
 {
 	zclient_send_reg_requests(zclient, VRF_DEFAULT);
+
+	static_fixup_vrf_ids(vrf_info_lookup(VRF_DEFAULT));
 }
 
-struct static_nht_data {
-	struct prefix *nh;
-
-	vrf_id_t nh_vrf_id;
-
-	uint32_t refcount;
-	uint8_t nh_num;
-};
-
 /* API to check whether the configured nexthop address is
  * one of its local connected address or not.
  */
@@ -165,14 +191,10 @@ static bool
 static_nexthop_is_local(vrf_id_t vrfid, struct prefix *addr, int family)
 {
 	if (family == AF_INET) {
-		if (if_lookup_exact_address(&addr->u.prefix4,
-					AF_INET,
-					vrfid))
+		if (if_address_is_local(&addr->u.prefix4, AF_INET, vrfid))
 			return true;
 	} else if (family == AF_INET6) {
-		if (if_lookup_exact_address(&addr->u.prefix6,
-					AF_INET6,
-					vrfid))
+		if (if_address_is_local(&addr->u.prefix6, AF_INET6, vrfid))
 			return true;
 	}
 	return false;
@@ -181,34 +203,37 @@ static int static_zebra_nexthop_update(ZAPI_CALLBACK_ARGS)
 {
 	struct static_nht_data *nhtd, lookup;
 	struct zapi_route nhr;
+	struct prefix matched;
 	afi_t afi = AFI_IP;
 
-	if (!zapi_nexthop_update_decode(zclient->ibuf, &nhr)) {
+	if (!zapi_nexthop_update_decode(zclient->ibuf, &matched, &nhr)) {
 		zlog_err("Failure to decode nexthop update message");
 		return 1;
 	}
 
-	if (nhr.prefix.family == AF_INET6)
+	if (matched.family == AF_INET6)
 		afi = AFI_IP6;
 
 	if (nhr.type == ZEBRA_ROUTE_CONNECT) {
-		if (static_nexthop_is_local(vrf_id, &nhr.prefix,
-					nhr.prefix.family))
+		if (static_nexthop_is_local(vrf_id, &matched,
+					    nhr.prefix.family))
 			nhr.nexthop_num = 0;
 	}
 
 	memset(&lookup, 0, sizeof(lookup));
-	lookup.nh = &nhr.prefix;
+	lookup.nh = matched;
 	lookup.nh_vrf_id = vrf_id;
+	lookup.safi = nhr.safi;
 
-	nhtd = hash_lookup(static_nht_hash, &lookup);
+	nhtd = static_nht_hash_find(static_nht_hash, &lookup);
 
 	if (nhtd) {
 		nhtd->nh_num = nhr.nexthop_num;
 
-		static_nht_reset_start(&nhr.prefix, afi, nhtd->nh_vrf_id);
-		static_nht_update(NULL, &nhr.prefix, nhr.nexthop_num, afi,
-				  nhtd->nh_vrf_id);
+		static_nht_reset_start(&matched, afi, nhr.safi,
+				       nhtd->nh_vrf_id);
+		static_nht_update(NULL, &matched, nhr.nexthop_num, afi,
+				  nhr.safi, nhtd->nh_vrf_id);
 	} else
 		zlog_err("No nhtd?");
 
@@ -218,190 +243,174 @@ static int static_zebra_nexthop_update(ZAPI_CALLBACK_ARGS)
 static void static_zebra_capabilities(struct zclient_capabilities *cap)
 {
 	mpls_enabled = cap->mpls_enabled;
+	zebra_ecmp_count = cap->ecmp;
 }
 
-static unsigned int static_nht_hash_key(const void *data)
+static struct static_nht_data *
+static_nht_hash_getref(const struct static_nht_data *ref)
 {
-	const struct static_nht_data *nhtd = data;
-	unsigned int key = 0;
+	struct static_nht_data *nhtd;
 
-	key = prefix_hash_key(nhtd->nh);
-	return jhash_1word(nhtd->nh_vrf_id, key);
+	nhtd = static_nht_hash_find(static_nht_hash, ref);
+	if (!nhtd) {
+		nhtd = XCALLOC(MTYPE_STATIC_NHT_DATA, sizeof(*nhtd));
+
+		prefix_copy(&nhtd->nh, &ref->nh);
+		nhtd->nh_vrf_id = ref->nh_vrf_id;
+		nhtd->safi = ref->safi;
+
+		static_nht_hash_add(static_nht_hash, nhtd);
+	}
+
+	nhtd->refcount++;
+	return nhtd;
 }
 
-static bool static_nht_hash_cmp(const void *d1, const void *d2)
+static bool static_nht_hash_decref(struct static_nht_data **nhtd_p)
 {
-	const struct static_nht_data *nhtd1 = d1;
-	const struct static_nht_data *nhtd2 = d2;
+	struct static_nht_data *nhtd = *nhtd_p;
 
-	if (nhtd1->nh_vrf_id != nhtd2->nh_vrf_id)
-		return false;
+	*nhtd_p = NULL;
 
-	return prefix_same(nhtd1->nh, nhtd2->nh);
+	if (--nhtd->refcount > 0)
+		return true;
+
+	static_nht_hash_del(static_nht_hash, nhtd);
+	XFREE(MTYPE_STATIC_NHT_DATA, nhtd);
+	return false;
 }
 
-static void *static_nht_hash_alloc(void *data)
+static void static_nht_hash_clear(void)
 {
-	struct static_nht_data *copy = data;
-	struct static_nht_data *new;
+	struct static_nht_data *nhtd;
 
-	new = XMALLOC(MTYPE_TMP, sizeof(*new));
-
-	new->nh = prefix_new();
-	prefix_copy(new->nh, copy->nh);
-	new->refcount = 0;
-	new->nh_num = 0;
-	new->nh_vrf_id = copy->nh_vrf_id;
-
-	return new;
+	while ((nhtd = static_nht_hash_pop(static_nht_hash)))
+		XFREE(MTYPE_STATIC_NHT_DATA, nhtd);
 }
 
-static void static_nht_hash_free(void *data)
+static bool static_zebra_nht_get_prefix(const struct static_nexthop *nh,
+					struct prefix *p)
 {
-	struct static_nht_data *nhtd = data;
-
-	prefix_free(&nhtd->nh);
-	XFREE(MTYPE_TMP, nhtd);
-}
-
-void static_zebra_nht_register(struct route_node *rn, struct static_nexthop *nh,
-			       bool reg)
-{
-	struct static_nht_data *nhtd, lookup;
-	uint32_t cmd;
-	struct prefix p;
-	afi_t afi = AFI_IP;
-
-	cmd = (reg) ?
-		ZEBRA_NEXTHOP_REGISTER : ZEBRA_NEXTHOP_UNREGISTER;
-
-	if (nh->nh_registered && reg)
-		return;
-
-	if (!nh->nh_registered && !reg)
-		return;
-
-	memset(&p, 0, sizeof(p));
 	switch (nh->type) {
 	case STATIC_IFNAME:
 	case STATIC_BLACKHOLE:
-		return;
+		p->family = AF_UNSPEC;
+		return false;
+
 	case STATIC_IPV4_GATEWAY:
 	case STATIC_IPV4_GATEWAY_IFNAME:
-		p.family = AF_INET;
-		p.prefixlen = IPV4_MAX_BITLEN;
-		p.u.prefix4 = nh->addr.ipv4;
-		afi = AFI_IP;
-		break;
+		p->family = AF_INET;
+		p->prefixlen = IPV4_MAX_BITLEN;
+		p->u.prefix4 = nh->addr.ipv4;
+		return true;
+
 	case STATIC_IPV6_GATEWAY:
 	case STATIC_IPV6_GATEWAY_IFNAME:
-		p.family = AF_INET6;
-		p.prefixlen = IPV6_MAX_BITLEN;
-		p.u.prefix6 = nh->addr.ipv6;
-		afi = AFI_IP6;
-		break;
+		p->family = AF_INET6;
+		p->prefixlen = IPV6_MAX_BITLEN;
+		p->u.prefix6 = nh->addr.ipv6;
+		return true;
 	}
 
-	memset(&lookup, 0, sizeof(lookup));
-	lookup.nh = &p;
+	assertf(0, "BUG: someone forgot to add nexthop type %u", nh->type);
+}
+
+void static_zebra_nht_register(struct static_nexthop *nh, bool reg)
+{
+	struct static_path *pn = nh->pn;
+	struct route_node *rn = pn->rn;
+	struct static_route_info *si = static_route_info_from_rnode(rn);
+	struct static_nht_data *nhtd, lookup = {};
+	uint32_t cmd;
+
+	if (!static_zebra_nht_get_prefix(nh, &lookup.nh))
+		return;
 	lookup.nh_vrf_id = nh->nh_vrf_id;
+	lookup.safi = si->safi;
+
+	if (nh->nh_registered) {
+		/* nh->nh_registered means we own a reference on the nhtd */
+		nhtd = static_nht_hash_find(static_nht_hash, &lookup);
+
+		assertf(nhtd, "BUG: NH %pFX registered but not in hashtable",
+			&lookup.nh);
+	} else if (reg) {
+		nhtd = static_nht_hash_getref(&lookup);
+
+		if (nhtd->refcount > 1)
+			DEBUGD(&static_dbg_route,
+			       "Reusing registered nexthop(%pFX) for %pRN %d",
+			       &lookup.nh, rn, nhtd->nh_num);
+	} else {
+		/* !reg && !nh->nh_registered */
+		zlog_warn("trying to unregister nexthop %pFX twice",
+			  &lookup.nh);
+		return;
+	}
 
 	nh->nh_registered = reg;
 
 	if (reg) {
-		nhtd = hash_get(static_nht_hash, &lookup,
-				static_nht_hash_alloc);
-		nhtd->refcount++;
+		if (nhtd->nh_num) {
+			/* refresh with existing data */
+			afi_t afi = prefix_afi(&lookup.nh);
 
-		if (debug)
-			zlog_debug("Registered nexthop(%pFX) for %pRN %d", &p,
-				   rn, nhtd->nh_num);
-		if (nhtd->refcount > 1 && nhtd->nh_num) {
-			static_nht_update(&rn->p, nhtd->nh, nhtd->nh_num, afi,
-					  nh->nh_vrf_id);
+			if (nh->state == STATIC_NOT_INSTALLED)
+				nh->state = STATIC_START;
+			static_nht_update(&rn->p, &nhtd->nh, nhtd->nh_num, afi,
+					  si->safi, nh->nh_vrf_id);
 			return;
 		}
+
+		if (nhtd->registered)
+			/* have no data, but did send register */
+			return;
+
+		cmd = ZEBRA_NEXTHOP_REGISTER;
+		DEBUGD(&static_dbg_route, "Registering nexthop(%pFX) for %pRN",
+		       &lookup.nh, rn);
 	} else {
-		nhtd = hash_lookup(static_nht_hash, &lookup);
-		if (!nhtd)
+		bool was_zebra_registered;
+
+		was_zebra_registered = nhtd->registered;
+		if (static_nht_hash_decref(&nhtd))
+			/* still got references alive */
 			return;
 
-		nhtd->refcount--;
-		if (nhtd->refcount >= 1)
+		/* NB: nhtd is now NULL. */
+		if (!was_zebra_registered)
 			return;
 
-		hash_release(static_nht_hash, nhtd);
-		static_nht_hash_free(nhtd);
+		cmd = ZEBRA_NEXTHOP_UNREGISTER;
+		DEBUGD(&static_dbg_route,
+		       "Unregistering nexthop(%pFX) for %pRN", &lookup.nh, rn);
 	}
 
-	if (zclient_send_rnh(zclient, cmd, &p, false, nh->nh_vrf_id) < 0)
-		zlog_warn("%s: Failure to send nexthop to zebra", __func__);
-}
-/*
- * When nexthop gets updated via configuration then use the
- * already registered NH and resend the route to zebra
- */
-int static_zebra_nh_update(struct route_node *rn, struct static_nexthop *nh)
-{
-	struct static_nht_data *nhtd, lookup = {};
-	struct prefix p = {};
-	afi_t afi = AFI_IP;
-
-	if (!nh->nh_registered)
-		return 0;
-
-	switch (nh->type) {
-	case STATIC_IFNAME:
-	case STATIC_BLACKHOLE:
-		return 0;
-	case STATIC_IPV4_GATEWAY:
-	case STATIC_IPV4_GATEWAY_IFNAME:
-		p.family = AF_INET;
-		p.prefixlen = IPV4_MAX_BITLEN;
-		p.u.prefix4 = nh->addr.ipv4;
-		afi = AFI_IP;
-		break;
-	case STATIC_IPV6_GATEWAY:
-	case STATIC_IPV6_GATEWAY_IFNAME:
-		p.family = AF_INET6;
-		p.prefixlen = IPV6_MAX_BITLEN;
-		p.u.prefix6 = nh->addr.ipv6;
-		afi = AFI_IP6;
-		break;
-	}
-
-	lookup.nh = &p;
-	lookup.nh_vrf_id = nh->nh_vrf_id;
-
-	nhtd = hash_lookup(static_nht_hash, &lookup);
-	if (nhtd && nhtd->nh_num) {
-		nh->state = STATIC_START;
-		static_nht_update(&rn->p, nhtd->nh, nhtd->nh_num, afi,
-				  nh->nh_vrf_id);
-		return 1;
-	}
-	return 0;
+	if (zclient_send_rnh(zclient, cmd, &lookup.nh, si->safi, false, false,
+			     nh->nh_vrf_id) == ZCLIENT_SEND_FAILURE)
+		zlog_warn("%s: Failure to send nexthop %pFX for %pRN to zebra",
+			  __func__, &lookup.nh, rn);
+	else if (reg)
+		nhtd->registered = true;
 }
 
-extern void static_zebra_route_add(struct route_node *rn,
-				   struct static_path *pn, safi_t safi,
-				   bool install)
+extern void static_zebra_route_add(struct static_path *pn, bool install)
 {
+	struct route_node *rn = pn->rn;
+	struct static_route_info *si = rn->info;
 	struct static_nexthop *nh;
 	const struct prefix *p, *src_pp;
 	struct zapi_nexthop *api_nh;
 	struct zapi_route api;
 	uint32_t nh_num = 0;
-	struct stable_info *info;
 
 	p = src_pp = NULL;
 	srcdest_rnode_prefixes(rn, &p, &src_pp);
 
 	memset(&api, 0, sizeof(api));
-	info = static_get_stable_info(rn);
-	api.vrf_id = GET_STABLE_VRF_ID(info);
+	api.vrf_id = si->svrf->vrf->vrf_id;
 	api.type = ZEBRA_ROUTE_STATIC;
-	api.safi = safi;
+	api.safi = si->safi;
 	memcpy(&api.prefix, p, sizeof(api.prefix));
 
 	if (src_pp) {
@@ -424,6 +433,10 @@ extern void static_zebra_route_add(struct route_node *rn,
 		api.tableid = pn->table_id;
 	}
 	frr_each(static_nexthop_list, &pn->nexthop_list, nh) {
+		/* Don't overrun the nexthop array */
+		if (nh_num == zebra_ecmp_count)
+			break;
+
 		api_nh = &api.nexthops[nh_num];
 		if (nh->nh_vrf_id == VRF_UNKNOWN)
 			continue;
@@ -509,6 +522,13 @@ extern void static_zebra_route_add(struct route_node *rn,
 			   zclient, &api);
 }
 
+static zclient_handler *const static_handlers[] = {
+	[ZEBRA_INTERFACE_ADDRESS_ADD] = interface_address_add,
+	[ZEBRA_INTERFACE_ADDRESS_DELETE] = interface_address_delete,
+	[ZEBRA_ROUTE_NOTIFY_OWNER] = route_notify_owner,
+	[ZEBRA_NEXTHOP_UPDATE] = static_zebra_nexthop_update,
+};
+
 void static_zebra_init(void)
 {
 	struct zclient_options opt = { .receive_notify = true };
@@ -516,19 +536,27 @@ void static_zebra_init(void)
 	if_zapi_callbacks(static_ifp_create, static_ifp_up,
 			  static_ifp_down, static_ifp_destroy);
 
-	zclient = zclient_new(master, &opt);
+	zclient = zclient_new(master, &opt, static_handlers,
+			      array_size(static_handlers));
 
 	zclient_init(zclient, ZEBRA_ROUTE_STATIC, 0, &static_privs);
 	zclient->zebra_capabilities = static_zebra_capabilities;
 	zclient->zebra_connected = zebra_connected;
-	zclient->interface_address_add = interface_address_add;
-	zclient->interface_address_delete = interface_address_delete;
-	zclient->route_notify_owner = route_notify_owner;
-	zclient->nexthop_update = static_zebra_nexthop_update;
 
-	static_nht_hash = hash_create(static_nht_hash_key,
-				      static_nht_hash_cmp,
-				      "Static Nexthop Tracking hash");
+	static_nht_hash_init(static_nht_hash);
+}
+
+/* static_zebra_stop used by tests/lib/test_grpc.cpp */
+void static_zebra_stop(void)
+{
+	static_nht_hash_clear();
+	static_nht_hash_fini(static_nht_hash);
+
+	if (!zclient)
+		return;
+	zclient_stop(zclient);
+	zclient_free(zclient);
+	zclient = NULL;
 }
 
 void static_zebra_vrf_register(struct vrf *vrf)
